
from .logger import get_logger
import numpy as np
from scipy.stats import gmean
import pandas as pd
import math


class ReactionExp():
    """
        This class is used for reaction expression analysis
    """

    def __init__(self,gem,gene_reg={},gene_exp={}):
        ## gene_reg={'gene_id':[],'gene_fold_chnage':[],'up':value (default 2),'down':value (default 2)}
        ## gene_reg={'gene_id':[],'gene_expression':[],'top_percent':value (default 15%),'bottom_percent': value (default 15%)}
        analysis=self.test_type_of_analysis(gene_reg,gene_exp)
        if 'reg' in analysis.keys() and analysis['reg']=='yes':
            # apply reaction regulation
            up_rxns,down_rxns,rxn_df=self.getReactionRegulation(gem,gene_reg)
            self.up_rxns=up_rxns
            self.down_rxns=down_rxns
            self.rxn_df=rxn_df
        elif 'context' in analysis.keys() and analysis['context']=='yes':
            high_rxns,low_rxns,rxn_df=self.getReactionExpression(gem,gene_exp)
            self.high_rxns=high_rxns
            self.low_rxns=low_rxns
            self.rxn_df=rxn_df

    def test_type_of_analysis(self,gene_reg,gene_exp):
        ''' set parameters'''
        analysis={}
        logger = get_logger('expression_logger')
        self.logger=logger
        if bool(gene_reg):
            self.logger=logger.info('start analysis of comparison between conditions .....')
            ## now we want to check gene_id and values
            cond1=('gene_id' in gene_reg.keys())
            cond2=('fold_change' in gene_reg.keys())
            if  cond1 and  cond2 and len(gene_reg['gene_id'])>0 and len(gene_reg['fold_change'])>0:
                analysis={'reg':'yes'}
            else:
                self.logger.error('problem in input file please prepare as condition.txt')
        elif bool(gene_exp):
            self.logger.info('start analysis of context (tissue-specific).....')

            cond1=('gene_id' in gene_exp.keys())
            cond2=('exp_val' in gene_exp.keys())
            if  cond1 and  cond2 and len(gene_exp['gene_id'])>0 and len(gene_exp['exp_val'])>0:
                analysis={'context':'yes'}
            else:
                self.logger.error('problem in input file please prepare as context.txt')


        elif (bool(gene_reg)) and ( bool(gene_reg)):
            self.logger.info('Asking for both context (tissue-specific) and conditions' + \
            'comparison at the same time.....\n please pass only one (either gene_reg or gene_exp)')
        else:
            self.logger.info('You do not want to do anay analysis...\n( please pass only one (either gene_reg or gene_exp))')
        return analysis

    def getReactionRegulation(self,gem,gene_reg):
        ''' apply gene protein reaction rule for gene expression '''
        # get list of gpr and list of reactions ids
        # set by default parameters
        if 'up_cutoff' not in gene_reg.keys():
            gene_reg['up_cutoff']=2
        if 'down_cutoff' not in gene_reg.keys():
            gene_reg['down_cutoff']=2
        gpr_assoicated_reaction=[]
        gprs=[]

        for rxn in gem.reactions:
        	if rxn.gene_reaction_rule!="":
        		gpr_assoicated_reaction.append(rxn)
        		gprs.append(rxn.gene_reaction_rule)
        if len(gprs)>0:
            gene_reg_dict=dict(zip(gene_reg['gene_id'],gene_reg['fold_change']))
            reaction_regulation=self.evalGPR(gprs,gpr_assoicated_reaction,gene_reg_dict,OR_str='mean',AND_str='geomean')
            # get not nan values
            reaction_regulation={k.id: v for k, v in reaction_regulation.items() if pd.Series(v).notna().all()}
            reaction_df = pd.DataFrame(list(reaction_regulation.items()),columns = ['Reaction id','Fold change'])
            sorted_reaction_rgulation_df=reaction_df.sort_values(by=['Fold change'],ascending=False)
            if not (sorted_reaction_rgulation_df.empty):
                up_rxns,down_rxns=self.applyCutoffOnRxndf(sorted_reaction_rgulation_df,gene_reg['up_cutoff'],gene_reg['down_cutoff'],colname='Fold change')
                return up_rxns,down_rxns,sorted_reaction_rgulation_df
            else:
                self.logger.error('Can not map gpr to values. please check gpr rules ')

        else:
            self.logger.error('In the model there is no gene protein reaction rule is associated')


    def getReactionExpression(self,gem,gene_exp):
        ''' apply gene protein reaction rule for gene expression '''
        # get list of gpr and list of reactions ids
        # set by default parameters
        if 'high_cutoff' not in gene_exp.keys():
            gene_exp['high_cutoff']=.15
        if 'low_cutoff' not in gene_exp.keys():
            gene_exp['low_cutoff']=.15
        gpr_assoicated_reaction=[]
        gprs=[]

        for rxn in gem.reactions:
        	if rxn.gene_reaction_rule!="":
        		gpr_assoicated_reaction.append(rxn)
        		gprs.append(rxn.gene_reaction_rule)
        if len(gprs)>0:
            gene_exp_dict=dict(zip(gene_exp['gene_id'],gene_exp['exp_val']))
            reaction_regulation=self.evalGPR(gprs,gpr_assoicated_reaction,gene_exp_dict,OR_str='sum',AND_str='min')
            # get not nan values
            reaction_regulation={k.id: v for k, v in reaction_regulation.items() if pd.Series(v).notna().all()}
            reaction_df = pd.DataFrame(list(reaction_regulation.items()),columns = ['Reaction id','Expression value'])
            sorted_reaction_rgulation_df=reaction_df.sort_values(by=['Expression value'],ascending=False)
            if not (sorted_reaction_rgulation_df.empty):
                up_rxns,down_rxns=self.applyHighLowCutoff(sorted_reaction_rgulation_df,gene_exp['high_cutoff'],gene_exp['low_cutoff'],colname='Expression value')
                return up_rxns,down_rxns,sorted_reaction_rgulation_df
            else:
                self.logger.error('Can not map gpr to values. please check gpr rules ')

        else:
            self.logger.error('In the model there is no gene protein reaction rule is associated')


    def applyCutoffOnRxndf(self,rxn_df,up_cut,down_cut,colname):
        ''' apply cutoff on reaction regulations '''

        up_df=rxn_df[rxn_df[colname]>=up_cut]
        down_df=rxn_df[rxn_df[colname]<=down_cut]
        up_rxns=up_df['Reaction id'].to_list()
        down_rxns=down_df['Reaction id'].to_list()
        return up_rxns,down_rxns


    def applyHighLowCutoff(self,rxn_df,up_cut,down_cut,colname):
        ''' apply cutoff on reaction regulations '''
        up_idx=[rxn_df.index[i] for i in range(math.floor(rxn_df.shape[0]*up_cut))]

        down_idx=[rxn_df.index[rxn_df.shape[0]-1-i] for i in range(math.floor(rxn_df.shape[0]*down_cut))]

        up_df=rxn_df.loc[up_idx,:].copy()
        down_df=rxn_df.loc[down_idx,:].copy()
        up_rxns=up_df['Reaction id'].to_list()
        down_rxns=down_df['Reaction id'].to_list()
        return up_rxns,down_rxns


    def evalGPR(self,gpr_id,reaction_id,gene_reg_dict,OR_str='max',AND_str='min'):
        ''' This function is written for evaluate GPR. OR and AND is replaces by function such as add, min'''
        #### INPUT ###
        # gpr_id ## list of gpr rules
        # reaction_id ## list of reaction ids
        # gene_df is DataFrame: genes and columns as values
        # OR_str='max'(by default). you can choose 'mean','sum'
        # AND_str='min', (by default). you can choose 'geomean'
        #### OUTPUT ###

        # if ('genes' in gene_df.columns) and ('exp_vals' in gene_df.columns):
        #     gene_reg_dict=dict(zip(gene_df['genes'],gene_df['exp_vals']))
        # else:
        #     print(''' Please give input for genes and their values.
        #      "genes" and "exp_vals" should be header for the sample_gene_expression file (Which is tab delimited).
        #      The genes column contains model genes and the exp_vals column contains expression measurement ''')
        #     exit()

        i=0
        varying_reaction={}
        for ids in gpr_id:
            reaction=reaction_id[i]
            i=i+1
            ids=ids.strip("\n")
            or_part=ids.split(" or ")

            or_vals=[]
            for word in or_part:
                and_part=word.split(" and ")
                ### if length of and_part is more than 1 means (and) is the part of the string
                and_vals=[]
                for rv in and_part:
                    rv=rv.strip()
                    rv=rv.strip("(")
                    rv=rv.strip(")")
                    if rv in gene_reg_dict.keys() and (not (gene_reg_dict[rv]==np.nan)):
                        and_vals.append(gene_reg_dict[rv])

                if len(and_vals)>0:
                    if AND_str=='min':
                        or_vals.append(np.min(and_vals))
                    elif AND_str=='geomean':
                        or_vals.append(gmean(and_vals))
                    else:
                        print("Put 'min' or 'geomean' for  AND string ")
                else:
                    print(" gene value is not found or nan for the reaction = %s" %reaction.id)
                    continue

            if len(or_vals)>0:
                ## now we ar going to apply sum rules
                if OR_str=='max':
                    final_val=np.max(or_vals)
                elif OR_str=='mean':
                    final_val=np.mean(or_vals)
                elif OR_str=='sum':
                    final_val=np.sum(or_vals)
                else:
                    print("Put 'max' or 'mean' or 'sum' for  AND string ")

            else:
                print('why or_vals are empty')
                final_val=np.nan

            varying_reaction[reaction]=final_val
        return varying_reaction



# -*- coding: utf-8 -*-
"""
.. module:: MiNEApy
   :platform: Unix, Mac
   :synopsis:Minimum network enrichment analysis

.. moduleauthor:: Vikash Pandey

Constraints declarations

"""
from pytfa.optim import ReactionConstraint, GenericConstraint, \
    ModelConstraint, GeneConstraint

class CatalyticConstraint(ReactionConstraint):
    """
    Class to represent a enzymatic constraint
    """

    prefix = 'CC_'


class SumFlux(GenericConstraint):
    """
    Class to represent a forbidden net flux directionality profile
    Looks like:
    F_rxn_1 + B_rxn_2 + ... + F_rxn_n <= n-1
    """

    def __init__(self, model, expr, id_, **kwargs):

        GenericConstraint.__init__(self,
                                   id_=id_,
                                   expr=expr,
                                   model=model,
                                   **kwargs)

    prefix = 'SumFlux_'



import numpy as np
# from cobra.flux_analysis import flux_variability_analysis
from pytfa.analysis.variability import variability_analysis

def remove_blocked_reactions(model):
    epsilon = model.solver.configuration.tolerances.feasibility
    # fva = flux_variability_analysis(model)
    fva = variability_analysis(model, kind='reactions')
    # Blocked reactions have max and min at 0
    df = fva[ (fva.max(axis=1).abs()<1*epsilon)
            & (fva.min(axis=1).abs()<1*epsilon)]
    rid_to_rm = df.index

    model.remove_reactions(rid_to_rm)

    return df


def trim_epsilon_mets(met_dict, epsilon):

    n = int(-1*np.log10(epsilon))
    round_dict = {x:np.round(v,n) for x,v in met_dict.items()}
    met_dict.update(round_dict)

    rm_list = [x for x,v in met_dict.items() if abs(v) <= epsilon]
    [met_dict.pop(x) for x in rm_list]

    return met_dict

def set_medium(model, medium_dict, inplace):
    if inplace:
        new = model
    else:
        new = model.copy()

    if medium_dict is None or not medium_dict:
        return new

    for rxn_id, lb in medium_dict.items():
        new.reactions.get_by_id(rxn_id).lower_bound = lb

    return new


# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis

.. moduleauthor:: pyTFA team

Thermodynamic cobra_model class and methods definition


"""

import re
from copy import deepcopy
from math import log

import pandas as pd
from cobra import Model

from pytfa.core.model import LCSBModel
from pytfa.thermo import std
# from .metabolite import MetaboliteThermo
# from .reaction import calcDGtpt_rhs, calcDGR_cues, get_debye_huckel_b
# from .utils import (
#     check_reaction_balance,
#     check_transport_reaction,
#     find_transported_mets,
#     get_reaction_compartment,
# )
from pytfa.optim.constraints import (
    SimultaneousUse,
    NegativeDeltaG,
    BackwardDeltaGCoupling,
    ForwardDeltaGCoupling,
    BackwardDirectionCoupling,
    ForwardDirectionCoupling,
    ReactionConstraint,
    MetaboliteConstraint,
    DisplacementCoupling,
)
from pytfa.optim.variables import (
    ThermoDisplacement,
    DeltaGstd,
    DeltaG,
    ForwardUseVariable,
    BackwardUseVariable,
    LogConcentration,
    ReactionVariable,
    MetaboliteVariable,
)
from pytfa.utils import numerics
from pytfa.utils.logger import get_bistream_logger
#
BIGM = numerics.BIGM
BIGM_THERMO = numerics.BIGM_THERMO
BIGM_DG = numerics.BIGM_DG
BIGM_P = numerics.BIGM_P
EPSILON = numerics.EPSILON
MAX_STOICH = 10


class ThermoModel_WithoutInfo(LCSBModel, Model):
    """
    A class representing a cobra_model with thermodynamics information

    """

    def __init__(self, model=Model(), name=None):

        """
        :param float temperature: the temperature (K) at which to perform the calculations
        :param dict thermo_data: The thermodynamic database
        :type temperature: float
        """

        LCSBModel.__init__(self, model, name)

        self.logger = get_bistream_logger('ME model' + str(self.name))
        #self._init_thermo()

        self.logger.info('# Model initialized ')

    def _convert_reaction(self, rxn, verbose):
        """

        :param rxn:
        :param verbose:
        :return:
        """
        FU_rxn = self.add_variable(ForwardUseVariable, rxn)
        BU_rxn = self.add_variable(BackwardUseVariable, rxn)

        # create the prevent simultaneous use constraints
        # SU_rxn: FU_rxn + BU_rxn <= 1
        CLHS = FU_rxn + BU_rxn
        self.add_constraint(SimultaneousUse, rxn, CLHS, ub=1)

        # create constraints that control fluxes with their use variables
        # UF_rxn: F_rxn - M FU_rxn < 0
        F_rxn = rxn.forward_variable
        CLHS = F_rxn - FU_rxn * BIGM
        self.add_constraint(ForwardDirectionCoupling, rxn, CLHS, ub=0)

        # UR_rxn: R_rxn - M RU_rxn < 0
        R_rxn = rxn.reverse_variable
        CLHS = R_rxn - BU_rxn * BIGM
        self.add_constraint(BackwardDirectionCoupling, rxn, CLHS, ub=0)

    def convert_withoutInfo(
        self, verbose=True
    ):
        """ Converts a cobra_model into a tFBA ready cobra_model by adding the
        thermodynamic constraints required

        .. warning::
            This function requires you to have already called
            :func:`~.pytfa.ThermoModel.prepare`, otherwise it will raise an Exception !

        """

        self.logger.info("# Model conversion starting...")

        ###########################################
        # CONSTANTS & PARAMETERS for tFBA problem #
        ###########################################

        # value for the bigM in big M constraints such as:
        # UF_rxn: F_rxn - M*FU_rxn < 0
        bigM = BIGM
        # Check each reactions' bounds
        for reaction in self.reactions:
            if (
                reaction.lower_bound < -bigM - EPSILON
                or reaction.upper_bound > bigM + EPSILON
            ):
                raise Exception("flux bounds too wide or big M not big enough")
            if reaction.lower_bound < -bigM:
                reaction.lower_bound = -bigM
            if reaction.upper_bound > bigM:
                reaction.upper_bound = bigM

        ###################
        # INPUTS & CHECKS #
        ###################

        ## For each reaction...
        for rxn in self.reactions:
            self._convert_reaction(
                rxn, verbose
            )

        # CONSISTENCY CHECKS

        # Creating the objective
        if len(self.objective.variables) == 0:
            self.logger.warning("Objective not found")

        self.logger.info("# Model conversion done.")
        self.logger.info("# Updating cobra_model variables...")
        self.repair()
        self.logger.info("# cobra_model variables are up-to-date")

    def print_info(self, specific=False):
        """
        Print information and counts for the cobra_model
        :return:
        """
        if not specific:
            LCSBModel.print_info(self)

        n_metabolites = len(self.metabolites)
        n_reactions = len(self.reactions)
        n_metabolites_thermo = len(
            # [
            #     x
            #     for x in self.metabolites
            #     if hasattr(x, "thermo") and x.thermo["id"]
            # ]
            self._var_kinds[LogConcentration.__name__]
        )
        n_reactions_thermo = len(
            # [
            #     x
            #     for x in self.reactions
            #     if x.id is not None
            #     and hasattr(x, "thermo")
            #     and x.thermo["computed"]
            # ]
            self._cons_kinds[ForwardDeltaGCoupling.__name__]
        )

        info = pd.DataFrame(columns=["value"])
        info.loc["num metabolites(thermo)"] = n_metabolites_thermo
        info.loc["num reactions(thermo)"] = n_reactions_thermo
        info.loc["pct metabolites(thermo)"] = (
            n_metabolites_thermo / n_metabolites * 100
        )
        info.loc["pct reactions(thermo)"] = (
            n_reactions_thermo / n_reactions * 100
        )
        info.index.name = "key"

        print(info)

    def __deepcopy__(self, memo):
        """

        :param memo:
        :return:
        """

        return self.copy()

    def copy(self):

        from ..io.dict import model_from_dict, model_to_dict
        from ..optim.utils import copy_solver_configuration

        dictmodel = model_to_dict(self)
        new = model_from_dict(dictmodel)

        copy_solver_configuration(self, new)

        return new



# -*- coding: utf-8 -*-
"""
.. module:: pytfa
   :platform: Unix, Windows
   :synopsis: Thermodynamics-based Flux Analysis
.. moduleauthor:: pyTFA team
Logging utilities
"""

import logging
import os
import time

LOGFOLDERNAME = 'logs'

def get_logger(name):
    """
    Sets up a logger that outputs INFO+ messages on stdout and DEBUG+ messages
    in the log file
    :param name: a class __name__ attribute
    :return:
    """
    logger = logging.getLogger(name)

    try:
        os.makedirs(LOGFOLDERNAME)
    except FileExistsError:
        pass

    if not logger.handlers:
        logger.setLevel(logging.DEBUG)

        # create a file handler
        filename = name+get_timestr()+'.log'
        path = os.path.join(LOGFOLDERNAME, filename)
        file_handler = logging.FileHandler(path)
        file_handler.setLevel(logging.DEBUG)

        # create a stream handler
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.INFO)

        # create a logging format
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        stream_handler.setFormatter(formatter)

        # add the handlers to the logger
        logger.addHandler(file_handler)
        logger.addHandler(stream_handler)

    return logger

def get_timestr():
    timestr = time.strftime("_%Y%m%d_%H%M%S")
    return timestr



from pytfa.analysis import  variability_analysis
from cobra import Reaction
#from cobra.flux_analysis.variability import flux_variability_analysis
from pytfa.optim.utils import symbol_sum
from pytfa.thermo.utils import is_exchange, check_transport_reaction
# from .utils import trim_epsilon_mets

from pytfa.optim.variables import ReactionVariable, BinaryVariable, get_binary_type
from pytfa.optim.constraints import ReactionConstraint, ForbiddenProfile

from numpy import sum, round,Inf

from optlang.interface import INFEASIBLE, TIME_LIMIT, OPTIMAL

from tqdm import tqdm

import pickle,os
import numpy as np

from collections import defaultdict, namedtuple
from .thermo_model import ThermoModel_WithoutInfo
from ..optim.constraints import SumFlux
from ..optim.variables import FluxSumVar

CPLEX = 'optlang-cplex'
GUROBI = 'optlang-gurobi'
GLPK = 'optlang-glpk'


# Transforms (OnePerBBB --> oneperbbb), (one_per_bbb --> oneperbbb), etc ...
disambiguate = lambda s:s.lower().replace('_','')

Lump = namedtuple('Lump', ['id_', 'metabolites', 'subnetwork', 'gene_reaction_rule'])

class InfeasibleExcept(Exception):
    def __init__(self, status, feasibility):
        self.status = status
        self.feasibility = feasibility


class TimeoutExcept(Exception):
    def __init__(self, time_limit):
        self.time_limit = time_limit


class FluxKO(ReactionVariable, BinaryVariable):
    prefix = 'KO_'

    def __init__(self, reaction, **kwargs):
        ReactionVariable.__init__(self, reaction,
                                  type=get_binary_type(),
                                  **kwargs)

# Define a new constraint type:
class UseOrKOInt(ReactionConstraint):
    prefix = 'UKI_'
# Define a new constraint type:
class UseOrKOFlux(ReactionConstraint):
    prefix = 'UKF_'



class EnumerateTask:
    """
        A class encapsulating the MiNEA algorithm
    """
    def __init__(self, tfa_model, params):
        """
        :type tfa_model: pytfa model

        :param biomass_rxns: list of biomass reactions
        :type biomass_rxns: [GEM.biomass_rxn.id]

        :param core_subsystems: list of Metabolic Tasks or BBB
        :type core_subsystems: [string]

        :param growth_rate: theoretical maximum specific growth rate in 1/hr units
        :type growth_rate: float

        :param timeout_limit: the maximum amount of time allowed to compute each optimization. Default is 3600s (1 hour)
        :type timeout_limit: float (seconds)
        """

        self._tfa_model = tfa_model

        self._param_dict = params
        self.init_params()
        # Set containing every Metabolic tasks
        # self._MTs = self.mts
        # Set containing every BBB reaction
        self._rBBB = list()
        # For each reaction

        if self.mt_bbb:
            for rxn in self._tfa_model.reactions:
                # If it's a BBB reaction
                #print(rxn.id,rxn.name)
                if rxn.id in self.biomass_rxns:
                    self._rBBB.append(rxn)
        # Growth rate
        self._growth_rate = self.growth_rate

        # TODO : solver choice
        # TODO default : solver du modele
        #self._solver = 'optlang-cplex'

        self._tfa_model.solver.configuration.timeout = self.timeout_limit
        print("Timeout limit is {}s".format(self.timeout_limit))

        try:
            fva=pickle.load(open(os.getcwd()+'/fva.pickle','rb'))
        except:
            self._tfa_model.logger.info(" Removing blocked reaction (that will help in optimization) ...")
            fva=self._get_blocked_reaction() # these are the blocked reactions

            pickle.dump(fva,open(os.getcwd()+'/fva.pickle','wb'))
        epsilon = self._tfa_model.solver.configuration.tolerances.feasibility
        blocked_df = fva[ (fva.max(axis=1).abs()<1*epsilon)
                & (fva.min(axis=1).abs()<1*epsilon)]
        self._blocked_reactions=blocked_df
        self._fva=fva

        # non_blocked_reactions
        not_blocked_rxns=[]
        for rxn in self._tfa_model.reactions:
            if rxn.id not in blocked_df.index.to_list():
                not_blocked_rxns.append(rxn)
        self._not_blocked_rxns=not_blocked_rxns

        ### remove bloked reactions
        ### initialize lb and ub from thermodynamic variability
        ### make FBA model

        # self._prune_and_make_FBA_model()
        # self._sinks=self._prepare_Mts()

        ## apply sum of fluxes
        # self._apply_sum_of_flux()
        #
        #
        #
        #
        #  # lumpgem binary variables to deactivate non-core reactions. The reaction is deactivated when the value of
        # # the variable is 1
        self._activation_vars = {rxn: self._tfa_model.add_variable(kind=FluxKO,
                                                                   hook=rxn,
                                                                   lb=0,
                                                                   ub=1,
                                                                   queue=False)
                                 for rxn in not_blocked_rxns}



        self._generate_usage_constraints()
        self._generate_objective()
        self._sinks=self._prepare_Mts()



    def _prune_FBA_model(self,tmp_model,bloked_rxn_ids):
        """
        Remove bloked reactions and make FBA model.
        working on reduced FBA model
        """
        tmp_model.remove_reactions(bloked_rxn_ids)
        return tmp_model


    def _prune_and_make_FBA_model(self):
        """
        Remove bloked reactions and make FBA model.
        working on reduced FBA model
        """
        #epsilon = self._tfa_model.solver.configuration.tolerances.feasibility
        rid_to_rm=self._blocked_reactions.index ## remove these reactions
        tmp_model=self._tfa_model.copy()
        tmp_model.remove_reactions(rid_to_rm)
        rid=[]
        for rxn in tmp_model.reactions:
            sink = tmp_model.reactions.get_by_id(rxn.id)
            # Activate reaction by setting its lower bound
            try:
                lb=self._fva.loc[rxn.id,'minimum']
                ub=self._fva.loc[rxn.id,'maximum']
            # get in fva
                sink.lower_bound=lb
                sink.upper_bound=ub
                rid.append(rxn.id)
            except:
                continue
        ## build FBA model


        tfa_model=ThermoModel_WithoutInfo(tmp_model)

        tfa_model.convert_withoutInfo()
        self._source_gem = tfa_model
        self._tfa_model = tfa_model
        self.logger = tfa_model.logger
        self.model_type='fba'
        self.orig_rid=rid


    def _apply_sum_of_flux(self):
        """
        we apply sum of flux as a variable and do flux variability analysis using optimum value as constraint
        """
        #epsilon = self._tfa_model.solver.configuration.tolerances.feasibility
        # get forward use and backward use variables
        forward_variables = [rxn.forward_variable  for rxn in self._source_gem.reactions]
        reverse_variables = [ rxn.reverse_variable for rxn in self._source_gem.reactions]
        #self._source_gem.reactions[0].forward_variable + self._source_gem.reactions[0].reverse_variable

        self._source_gem.add_variable(kind=FluxSumVar,hook=self._source_gem,id_='FluxsumVar',lb=0,ub=Inf)
        self._FluxsumVar=self._source_gem.variables[-1]
        expr = symbol_sum(forward_variables+reverse_variables)
        new_expr=expr-self._FluxsumVar
        self._source_gem.add_constraint(kind=SumFlux,hook = self._source_gem,id_ = 'SumFlux',expr = new_expr,lb=0,ub=0)

        self._source_gem.objective = self._source_gem.problem.Objective(self._FluxsumVar, direction='min')
        solution=self._source_gem.optimize()
        self._solution=solution
        if solution.status=='optimal':
            self._FluxsumVar.lb=solution.objective_value
            self._FluxsumVar.ub=solution.objective_value+0.1*solution.objective_value

    def _solve_sum_of_flux(self):
        """
        we apply sum of flux as a variable and do flux variability analysis using optimum value as constraint
        """
        #epsilon = self._tfa_model.solver.configuration.tolerances.feasibility
        # get forward use and backward use variables
        forward_variables = [rxn.forward_variable  for rxn in self._tfa_model.reactions]
        reverse_variables = [ rxn.reverse_variable for rxn in self._tfa_model.reactions]
        #self._source_gem.reactions[0].forward_variable + self._source_gem.reactions[0].reverse_variable

        self._tfa_model.add_variable(kind=FluxSumVar,hook=self._tfa_model,id_='FluxsumVar',lb=0,ub=Inf)
        self._FluxsumVar=self._tfa_model.variables[-1]
        expr = symbol_sum(forward_variables+reverse_variables)
        new_expr=expr-self._FluxsumVar
        self._tfa_model.add_constraint(kind=SumFlux,hook = self._tfa_model,id_ = 'SumFlux',expr = new_expr,lb=0,ub=0)

        self._tfa_model.objective = self._tfa_model.problem.Objective(self._FluxsumVar, direction='min')



    def _generate_usage_constraints(self):
        """
        Generate carbon intake related constraints for each non-core reaction
        For each reaction rxn : rxn.forward_variable + rxn.reverse_variable + activation_var * C_uptake < C_uptake
        """
        flux_methods = ['flux', 'fluxes', 'both']
        int_methods = ['int', 'integer', 'both']

        if self.constraint_method.lower() not in flux_methods + int_methods:
            raise ArgumentError('{} is not a correct constraint method. '
                                'Choose among [Flux, Integer, Both]. '
                                'If you do not know what to choose, go for Flux.'
                                'If it is too slow, go for integer.'
                                'If you get strange lumps, go for both'
                                .format(self.constraint_method))

        for rxn in self._not_blocked_rxns:
            activation_var = self._activation_vars[rxn]
            if self.constraint_method.lower() in flux_methods:
                bigM = 100
                reac_var = rxn.forward_variable + rxn.reverse_variable + activation_var * bigM
                # adding the constraint to the model
                self._tfa_model.add_constraint(kind=UseOrKOFlux,
                                               hook=rxn,
                                               expr=reac_var,
                                               ub=bigM,
                                               lb=0,
                                               queue=True)
            if self.constraint_method.lower() in int_methods:
                fu = self._tfa_model.forward_use_variable .get_by_id(rxn.id)
                bu = self._tfa_model.backward_use_variable.get_by_id(rxn.id)
                reac_var = fu + bu + activation_var
                # adding the constraint to the model
                self._tfa_model.add_constraint(kind=UseOrKOInt,
                                               hook=rxn,
                                               expr=reac_var,
                                               ub=1,
                                               lb=0,
                                               queue=True)


        # push constraints in one bulk (faster)
        self._tfa_model._push_queue()
        # refresh constraint fields
        self._tfa_model.repair()



    def _get_blocked_reaction(self):
        """
        We compute flux variability analysis. so that we could block those reactions
        """


        fva = variability_analysis(self._tfa_model, kind='reactions')



        return fva

    def init_params(self):
        self.biomass_rxns = self._param_dict["biomass_rxns"]

        self.growth_rate = self._param_dict["growth_rate"]

        self.timeout_limit = self._param_dict["timeout"]
        self.mts = self._param_dict["MTs"]
        self.mt_bbb = self._param_dict["MT_BBB"]
        self.constraint_method = self._param_dict["constraint_method"]
        self.model_type = self._param_dict["model_type"]
        self.cofactor_pairs = self._param_dict["cofactor_pairs"]
        self.diverge_min = self._param_dict["diverge_min"]

    def get_cofactor_adjusted_stoich(self,rxn):
        stoich_dict = {x.id:v for x,v in rxn.metabolites.items()}

        for a,b in self.cofactor_pairs:
            try:
                na = stoich_dict[a] # looks like -54 atp_c
                nb = stoich_dict[b] # looks like +53 adp_c

                n = na+nb # looks like -1

                if n == 0:
                    self._tfa_model.logger.warn(
                        'Cofactor pair {}/{} is equimolar in reaction {}'
                        .format(a,b,rxn.id))
                elif n > 0:
                    n = -n
                    self._tfa_model.logger.warn(
                        'Cofactor pair {}/{} looks inverted in reaction {}'
                        .format(a,b,rxn.id))

                stoich_dict[a] =  n # looks like 1
                stoich_dict[b] = -n # looks like -1
            except KeyError:
                pass
        return stoich_dict



    def _prepare_Mts(self):
        """
        For each BBB (reactant of the biomass reactions), generate a sink, i.e an unbalanced reaction BBB ->
        of which purpose is to enable the BBB to be output of the GEM
        :return: the dict {BBB: sink} containing every BBB (keys) and their associated sinks
        """
        all_sinks = {}
        print("Preparing metabolic tasks...")

        for bio_rxn in self._rBBB:
            stoich_dict = self.get_cofactor_adjusted_stoich(bio_rxn)
            for met in bio_rxn.reactants:
                stoech_coeff = stoich_dict[met.id]
                # stoech_coeff < 0 indicates that the metabolite is a reactant
                if (stoech_coeff < 0) and (met not in all_sinks.keys()):
                    sink = Reaction("MT_" + bio_rxn.id + "_" + met.id)
                    sink.name = "MT_" + bio_rxn.name + "_" + met.name
                    # Subsystem specific to BBB sinks
                    sink.subsystem = "MTs"

                    # A sink is simply a reaction which consumes the BBB
                    sink.add_metabolites({met: -1})

                    # The sinks will be activated later (cf compute_lumps), one at a time
                    # sink.knock_out()

                    # The stoechiometric coefficients will be used to define the lower bound of the sink,
                    # thus it must be stored
                    all_sinks[met] = (sink.id, -stoech_coeff)
                    self._tfa_model.add_reactions([sink])

                # reactant already seen
                elif stoech_coeff < 0:
                    # The BBB has already been associated to a sink, so we simply increase the bound of the sink
                    all_sinks[met][1] -= stoech_coeff

        # Must be called before changing the reaction.thermo['computed'] values

        for met in self._tfa_model.metabolites:
            ### get metabolite by id

            if met.id in self.mts:
                sink = Reaction("MT_"+ met.id)
                sink.name = "MT_" + met.name
                # Subsystem specific to BBB sinks
                sink.subsystem = "MTs"

                # A sink is simply a reaction which consumes the BBB
                sink.add_metabolites({met: -1})
                all_sinks[met] = (sink.id, 1)
                self._tfa_model.add_reactions([sink])

        if self.model_type=='tfa':
            self._tfa_model.prepare()
            # Must be called before optimization
            self._tfa_model.convert()
        else:
            self._tfa_model.convert_withoutInfo()


        # self._activation_vars = {rxn: self._tfa_model.add_variable(kind=FluxKO,
        #                                                            hook=rxn,
        #                                                            lb=0,
        #                                                            ub=1,
        #                                                            queue=False)
        #                          for rxn in self._tfa_model.reactions}
        # for ncrxn in self._rncore:
        #     ncrxn.thermo['computed'] = False

        return all_sinks


    def _generate_objective(self):
        """
        Generate and add the maximization objective : set as many activation variables as possible to 1
        When an activation variable is set to 1, the corresponding non-core reaction is deactivated
        """
        # Sum of binary variables to be maximized

        objective_sum = symbol_sum(list(self._activation_vars.values()))
        # Set the sum as the objective function
        self._tfa_model.objective = self._tfa_model.problem.Objective(objective_sum, direction='max')

    def compute_mins(self, force_solve=False, method='OnePerBBB'):
        """
        For each BBB (reactant of the biomass reaction), add the corresponding sink to the model, then optimize and
        lump the result into one lumped reaction
        :param force_solve: Indicates whether the computations must continue when one lumping yields a status "infeasible"
        :return: The dict {BBB: lump} containing every lumped reactions, associated to their BBBs
        """

        # # Must be called before optimization
        # self._tfa_model.convert()
        # self._tfa_model.objective_direction = 'min'

        epsilon = self._tfa_model.solver.configuration.tolerances.feasibility

        the_method = disambiguate(method)
        print('Min network method detected: {}'.format(the_method))

        # dict: {metabolite: lumped_reaction}
        lumps = {}

        self._tfa_model.objective_direction = 'max'

        sink_iter = tqdm(self._sinks.items(), desc = 'met')
        # import pdb; pdb.set_trace()
        for met_BBB, (sink_id, stoech_coeff) in sink_iter:

            # Cute stuff

            sink_iter.set_description('met={}'.format(met_BBB.id[:10]))
            sink_iter.refresh()

            sink = self._tfa_model.reactions.get_by_id(sink_id)
            # Activate reaction by setting its lower bound
            prev_lb = sink.lower_bound
            # min_prod = self._growth_rate * stoech_coeff
            # sink.lower_bound = min_prod - epsilon
            ### we are going to optimize
            self._tfa_model.objective = self._tfa_model.problem.Objective(sink.forward_variable, direction='max')
            n_da = self._tfa_model.slim_optimize()

            try:
                # Timeout reached
                if self._tfa_model.solver.status == TIME_LIMIT:
                    raise TimeoutExcept(self._tfa_model.solver.configuration.timeout)
                # Not optimal status -> infeasible
                elif self._tfa_model.solver.status != OPTIMAL:
                    raise InfeasibleExcept(self._tfa_model.solver.status,
                                           self._tfa_model.solver.configuration.tolerances.feasibility)

            except (TimeoutExcept, InfeasibleExcept) as err:
                print('Can not Produced {}'.format(sink_id))
                self.lumps = lumps
                return lumps

            sink.lower_bound = sink.flux*0.8
            self._generate_objective()
            if the_method == 'oneperbbb':
                this_lump = self._lump_one_per_bbb(met_BBB, sink, force_solve)
                lumped_reactions = [this_lump] if this_lump is not None else list()
            elif the_method.startswith('min+'):
                try:
                    p = int(the_method.replace('min+',''))
                except ValueError:
                    raise ValueError('Min+p method must have p as an integer')
                lumped_reactions = self._lump_min_plus_p(met_BBB, sink, p, force_solve)
            elif the_method.startswith('min'):
                lumped_reactions = self._lump_min_plus_p(met_BBB, sink, 0, force_solve)
            else:
                raise ValueError('Lumping method not recognized: {}. '
                                 'Valid methods are '
                                 'OnePerBBB, Min, Min+p, p natural integer'
                                 .format(the_method))


            if not lumped_reactions:
                continue

            lumps[met_BBB] = lumped_reactions
            # Deactivating reaction by setting both bounds to 0
            sink.lower_bound = prev_lb
            # sink.knock_out()

        self.lumps = lumps
        return lumps

    def _lump_one_per_bbb_mix(self, met_BBB, sink, force_solve):
        """

        :param met_BBB:
        :param sink:
        :param force_solve:
        :return:
        """
        n_da = self._tfa_model.slim_optimize()

        try:
            # Timeout reached
            if self._tfa_model.solver.status == TIME_LIMIT:
                raise TimeoutExcept(self._tfa_model.solver.configuration.timeout)
            # Not optimal status -> infeasible
            elif self._tfa_model.solver.status != OPTIMAL:
                raise InfeasibleExcept(self._tfa_model.solver.status,
                                       self._tfa_model.solver.configuration.tolerances.feasibility)

        except (TimeoutExcept, InfeasibleExcept) as err:
            # If the user want to continue anyway, suits him
            if force_solve:
                # Raise a warning

                self._solve_sum_of_flux()
                n_da = self._tfa_model.slim_optimize()
                try:
                    # Timeout reached
                    if self._tfa_model.solver.status == TIME_LIMIT:
                        raise TimeoutExcept(self._tfa_model.solver.configuration.timeout)
                        # Not optimal status -> infeasible
                    elif self._tfa_model.solver.status != OPTIMAL:
                        raise InfeasibleExcept(self._tfa_model.solver.status,
                                           self._tfa_model.solver.configuration.tolerances.feasibility)
                except (TimeoutExcept, InfeasibleExcept) as err:
                    return None

                lumped_reaction = self._build_lump_sum_flux(met_BBB, sink)
                return [lumped_reaction,'sum_flux']


            else:
                raise err

        print('Produced {}'.format(sink.flux),
              'with {0:.0f} reactions deactivated'.format(n_da))
        lumped_reaction = self._build_lump(met_BBB, sink)

        return lumped_reaction



    def _lump_one_per_bbb(self, met_BBB, sink, force_solve):
        """

        :param met_BBB:
        :param sink:
        :param force_solve:
        :return:
        """

        #n_da = self._tfa_model.slim_optimize()
        solution = self._tfa_model.optimize()
        n_da=solution.objective_value

        try:
            # Timeout reached
            if self._tfa_model.solver.status == TIME_LIMIT:
                raise TimeoutExcept(self._tfa_model.solver.configuration.timeout)
            # Not optimal status -> infeasible
            elif self._tfa_model.solver.status != OPTIMAL:
                raise InfeasibleExcept(self._tfa_model.solver.status,
                                       self._tfa_model.solver.configuration.tolerances.feasibility)

        except (InfeasibleExcept) as err:
            # If the user want to continue anyway, suits him
            if force_solve:
                # Raise a warning
                self._solve_sum_of_flux()
                n_da = self._tfa_model.slim_optimize()
                try:
                    # Timeout reached
                    if self._tfa_model.solver.status == TIME_LIMIT:
                        raise TimeoutExcept(self._tfa_model.solver.configuration.timeout)
                        # Not optimal status -> infeasible
                    elif self._tfa_model.solver.status != OPTIMAL:
                        raise InfeasibleExcept(self._tfa_model.solver.status,
                                           self._tfa_model.solver.configuration.tolerances.feasibility)
                except (TimeoutExcept, InfeasibleExcept) as err:
                    return None

                lumped_reaction = self._build_lump_sum_flux(met_BBB, sink)
                return [lumped_reaction,'sum_flux']


            else:
                raise err

        print('Produced {}'.format(sink.flux),
              'with {0:.0f} reactions deactivated'.format(n_da))
        lumped_reaction = self._build_lump(met_BBB, sink)

        return lumped_reaction

    def _build_lump(self, met_BBB, sink):
        """
        This function uses the current solution of self._tfa_model

        :param met_BBB:
        :param sink:
        :return:
        """

        epsilon_int = self._tfa_model.solver.configuration.tolerances.integrality
        epsilon_flux = self._tfa_model.solver.configuration.tolerances.feasibility

        sigma = sink.flux
        lump_dict = dict()

        if (sigma > epsilon_flux ):
            lump_dict[sink] = sigma
            for rxn in self._not_blocked_rxns:
                if (self._activation_vars[rxn].variable.primal < epsilon_int):
                    lump_dict[rxn] = rxn.flux
        else:
            self._tfa_model.logger.info('Metabolite {} is not able to produce'
                                        'by environment or external medium'.format(met_BBB.id))
        return lump_dict


    def _build_lump_sum_flux(self, met_BBB, sink):
        """
        This function uses the current solution of self._tfa_model

        :param met_BBB:
        :param sink:
        :return:
        """

        # import pdb; pdb.set_trace()
        epsilon_flux = self._tfa_model.solver.configuration.tolerances.feasibility
        # rxns=self._not_blocked_rxns+[sink]
        # fluxes=np.array([r.flux for r in rxns])
        # non_zero_fluxes=fluxes[abs(fluxes)>epsilon_flux]
        # active_rxns=rxns[abs(fluxes)>epsilon_flux]
        #
        # try:
        #     lump_dict = dict(zip(active_rxns, non_zero_fluxes))
        # except:
        #     import pdb; pdb.set_trace()

        sigma = sink.flux
        lump_dict = dict()
        if (sigma > epsilon_flux ):
            lump_dict[sink] = sigma
            for rxn in self._not_blocked_rxns:
                if (abs(rxn.flux)>epsilon_flux):
                    lump_dict[rxn] = rxn.flux
        else:
            self._tfa_model.logger.info('Metabolite {} is not able to produce'
                                        'by environment or external medium'.format(met_BBB.id))
        return lump_dict


    def _lump_min_plus_p(self, met_BBB, sink, p, force_solve):
        """

        :param met_BBB:
        :param sink:
        :param force_solve:
        :return:
        """

        epsilon = self._tfa_model.solver.configuration.tolerances.integrality

        try:
            max_lumps =self._param_dict['max_lumps_per_BBB']
        except KeyError:
            # TODO: Put a warning
            max_lumps=10

        lumps = list()

        with self._tfa_model as model:
            activation_vars = model.get_variables_of_type(FluxKO)

            # Solve a first time, obtain minimal subnet
            # import pdb; pdb.set_trace()
            model.slim_optimize()
            max_deactivated_rxns = model.objective.value

            # Add constraint forbidding subnets bigger than p
            expr = symbol_sum(activation_vars)

            # The lower bound is the max number of deactivated, minus p
            # Which allows activating the minimal number of reactions, plus p
            # lb = max_deactivated_rxns - p
            model.add_constraint(kind=ForbiddenProfile,
                                 hook = model,
                                 id_ = 'MAX_DEACT_{}'.format(met_BBB.id),
                                 expr = expr,
                                 lb=0,
                                 ub=len(model.reactions)
                                 # lb = lb,
                                 # ub = max_deactivated_rxns,
                                 )

            n_deactivated_reactions = max_deactivated_rxns

            # While loop, break on infeasibility
            while len(lumps)<max_lumps:

                try:
                    this_lump = self._lump_one_per_bbb(met_BBB, sink, force_solve)
                except (InfeasibleExcept, TimeoutExcept) as e:
                    if force_solve:
                        pass
                    elif len(lumps) == 0:
                        # No solution AND no lump found
                        raise e
                if model.solver.status != OPTIMAL:
                    break
                elif  isinstance(this_lump, list) and this_lump[1]=='sum_flux':
                    this_lump=this_lump[0]
                    lumps.append(this_lump)
                    return lumps
                elif this_lump is None:
                    # Since the solver is optimal, and we caught optim errors before,
                    # Then the BBB is simply produced in enough quantity by the core
                    break

                lumps.append(this_lump)

                # Add constraint forbidding the previous solution

                is_inactivated = [x for x in activation_vars
                               if abs(x.variable.primal-1) < 2*epsilon]

                expr = symbol_sum(is_inactivated)
                model.add_constraint(kind=ForbiddenProfile,
                                     hook = model,
                                     id_ = '{}_{}_{}'.format(met_BBB.id,
                                                             n_deactivated_reactions,
                                                             len(lumps)),
                                     expr = expr,
                                     lb = max_deactivated_rxns-p-self._param_dict['diverge_min'],
                                     ub = n_deactivated_reactions-self._param_dict['diverge_min'],
                                     )

        # TODO: Update of dynamic properties not handled yet
        # upon exiting context manager
        model.repair()
        return lumps

    def sum_reactions(rxn_dict, id_ = 'summed_reaction', epsilon = 1e-9):
        """
        Keys are reactions
        Values are their multiplicative coefficient
        """
        stoich = defaultdict(int)

        for rxn,flux in rxn_dict.items():
            for x, coeff in rxn.metabolites.items():
                stoich[x.id] += coeff * flux

        gpr = ') and ('.join(x.gene_reaction_rule for x in rxn_dict if x.gene_reaction_rule)

        gpr = ('(' + gpr + ')') if gpr else ''

        stoich = trim_epsilon_mets(stoich, epsilon=epsilon)

        new = Lump(id_ = id_,
                   metabolites = stoich,
                   subnetwork = {x.id:v for x,v in rxn_dict.items()},
                   gene_reaction_rule=gpr)

        return new



    def compute_mins_network(self, force_solve=False, method='OnePerBBB'):
        """
        For each BBB (reactant of the biomass reaction), add the corresponding sink to the model, then optimize and
        lump the result into one lumped reaction
        :param force_solve: Indicates whether the computations must continue when one lumping yields a status "infeasible"
        :return: The dict {BBB: lump} containing every lumped reactions, associated to their BBBs
        """

        # # Must be called before optimization
        # self._tfa_model.convert()
        # self._tfa_model.objective_direction = 'min'

        epsilon = self._tfa_model.solver.configuration.tolerances.feasibility

        # self._tfa_model_save=self._tfa_model
        the_method = disambiguate(method)
        print('Min network method detected: {}'.format(the_method))

        # dict: {metabolite: lumped_reaction}
        lumps = {}

        self._tfa_model.objective_direction = 'max'

        sink_iter = tqdm(self._sinks.items(), desc = 'met')

        for met_BBB, (sink_id, stoech_coeff) in sink_iter:

            # Cute stuff

            sink_iter.set_description('met={}'.format(met_BBB.id[:10]))
            sink_iter.refresh()

            sink = self._tfa_model.reactions.get_by_id(sink_id)
            # Activate reaction by setting its lower bound
            prev_lb = sink.lower_bound
            min_prod = self._growth_rate * stoech_coeff
            sink.lower_bound = min_prod - epsilon

            ## apply sum of flux
            print('sink_lower_bound',sink.lower_bound)
            self._apply_sum_of_flux()
            ## variability_analysis
            try:
                fva=pickle.load(open(os.getcwd()+'/'+sink_id+'fva.pickle','rb'))
            except:
                fva = variability_analysis(self._tfa_model, kind='reactions')
                pickle.dump(fva,open(os.getcwd()+'/'+sink_id+'fva.pickle','wb'))


            epsilon = self._tfa_model.solver.configuration.tolerances.feasibility
            blocked_df = fva[ (fva.max(axis=1).abs()<1*epsilon)
                    & (fva.min(axis=1).abs()<1*epsilon)]


            ## non_blocked_reactions
            not_blocked_rxns=[]
            for rxn in self._tfa_model.reactions:
                if rxn.id not in blocked_df.index.to_list():
                    not_blocked_rxns.append(rxn)
            self._not_blocked_rxns=not_blocked_rxns
            rids=set([ r.id for r in not_blocked_rxns])
            comm=list(set(self.orig_rid)&rids)
            add_var_rxns=[]
            for id in comm:
                add_var_rxns.append(self._tfa_model.reactions.get_by_id(id))

            # tmp_model=self._prune_FBA_model(self._tfa_model,blocked_df.index)

             # lumpgem binary variables to deactivate non-core reactions. The reaction is deactivated when the value of
            # the variable is 1
            self._activation_vars = {rxn: self._tfa_model.add_variable(kind=FluxKO,
                                                                       hook=rxn,
                                                                       lb=0,
                                                                       ub=1,
                                                                       queue=False)
                                     for rxn in add_var_rxns}

            print('Number of reaction:',len(add_var_rxns))
            self._generate_usage_constraints()
            self._generate_objective()
            # self._sinks=self._prepare_Mts()
            ### remove sum_of_flux bound
            self._FluxsumVar.lb=0
            self._FluxsumVar.ub=0

            if the_method == 'oneperbbb':
                this_lump = self._lump_one_per_bbb(met_BBB, sink, force_solve)
                lumped_reactions = [this_lump] if this_lump is not None else list()
            elif the_method.startswith('min+'):
                try:
                    p = int(the_method.replace('min+',''))
                except ValueError:
                    raise ValueError('Min+p method must have p as an integer')
                lumped_reactions = self._lump_min_plus_p(met_BBB, sink, p, force_solve)
            elif the_method.startswith('min'):
                lumped_reactions = self._lump_min_plus_p(met_BBB, sink, 0, force_solve)
            else:
                raise ValueError('Lumping method not recognized: {}. '
                                 'Valid methods are '
                                 'OnePerBBB, Min, Min+p, p natural integer'
                                 .format(the_method))


            if not lumped_reactions:
                continue

            lumps[met_BBB] = lumped_reactions
            # Deactivating reaction by setting both bounds to 0
            sink.lower_bound = prev_lb
            # sink.knock_out()

        self.lumps = lumps
        return lumps



from .enumTask import EnumerateTask
from .utils import remove_blocked_reactions, set_medium
import yaml
from .mhygepdf import mhygepdf
import statsmodels.stats.multitest as mtest
from scipy.stats import hypergeom
# from .logger import get_bistream_logger
from .numerics import BIGM,EPSILON

from cobra import Model
import pickle,os
import pandas as pd

from pytfa.optim.constraints import (
    SimultaneousUse,
    BackwardDirectionCoupling,
    ForwardDirectionCoupling,
)
from pytfa.optim.variables import (
    ForwardUseVariable,
    BackwardUseVariable,
)



from .thermo_model import ThermoModel_WithoutInfo

class TaskEnrichment():
    """
        A class to perform metabolic task enrichment analysis
    """

    def __init__(self,gem,parameters_path, params_rxn,inplace=False):

        # params_rxns={'up_rxns':reg_analysis.up_rxns,'down_rxns':reg_analysis.down_rxns, 'de_rxns':[de_rxns]}
        # params_rxns={'high_rxns':reg_analysis.up_rxns,'low_rxns':reg_analysis.down_rxns}
        self.read_parameters(parameters_path)
        # If inplace is True, no deepcopy is performed : the modifications are applied directly onto the gem
        prepared_gem = set_medium(gem, self.params['medium'], inplace)

        self._gem = prepared_gem
        self._params_rxn = params_rxn
        # This one is used to perform the lumping
        self._source_gem = prepared_gem.copy()
        ###


        ### test model id TFA or not
        if hasattr(self._source_gem , 'logger'):
            self.logger = self._gem.logger
            self._model_type='tfa'
        else:
            # self.logger = get_bistream_logger('FBA_model' + str(self._gem.id))
            ## we are going to add Forward and reverse use variable for the reactions
            tfa_model=ThermoModel_WithoutInfo(prepared_gem)

            tfa_model.convert_withoutInfo()
            self._source_gem = tfa_model
            self.logger = tfa_model.logger
            self._model_type='fba'

        ###
        self.fill_default_params()
        self.set_solver()


    def read_parameters(self, parameters_path):
        with open(parameters_path, 'r') as stream:
            try:
                self.params = yaml.safe_load(stream)
                print("Opened parameters file")
            except yaml.YAMLError as exc:
                print(exc)

    def fill_default_params(self):
        # If auto is activated, automatically extracts inorganics from the gem
        if "growth_rate" not in self.params or self.params["growth_rate"] == "auto":
            self.logger.info("Setting minimal growth rate to 95% of the TFA solution")
            obj_val = self._source_gem.slim_optimize()
            self.logger.info("Setting minimal growth rate to {}".format(obj_val))
            self.params["growth_rate"] = 0.95*obj_val
        if "force_solve" not in self.params:
            self.params["force_solve"] = False
        if "timeout" not in self.params:
            self.logger.info("Using default timeout : 3600s")
            self.params["timeout"] = 3600
        if "diverge_min" not in self.params:
            self.logger.info("Using default diverge_min : 1")
            self.params["diverge_min"] = 1
        if "feasibility" not in self.params:
            self.logger.info("Using default solver feasibility : 1e-9")
            self.params["feasibility"] = 1e-9
        else:
            # numbers like 1e-9 are detected as strings by yaml module
            # to enable their use, we cast them into floats
            try:
                self.params["feasibility"] = float(self.params["feasibility"])
            except ValueError as v:
                self.logger.error(v)

    def set_solver(self):
        if "solver" not in self.params or self.params["solver"].lower() == "auto":
            return None
        elif 'gurobi' in self.params["solver"].lower():
            solver = 'gurobi'
        elif 'cplex' in self.params["solver"].lower():
            solver = 'cplex'
        elif 'glpk' in self.params["solver"].lower():
            solver = 'glpk'
        else:
            solver = self.params["solver"]

        self._gem.solver = solver
        self._source_gem.solver = solver

    def get_enrichment_table_for_de(self,mins,de_rxns,de_col='deregulated'):
        ''' get enrichment of deregulated reactions (up or down regulated)'''
        mets_synthesis=[]
        alternatives=[]
        met_description=[]
        pvalues=[]
        probabilities=[]
        percent_de=[]
        rxns_num=[]
        list_de_in_min=[]
        list_rxn_in_min=[]
        gem=self._source_gem

        total_rxns=[]
        model_rxns=[]



        for rxn in gem.reactions:
            model_rxns.append(rxn.id)
            if rxn.gene_reaction_rule!="":
                total_rxns.append(rxn.id)
        total_rxns=set(total_rxns)
        model_rxns=set(model_rxns)
        without_gpr_rxns=[]

        for k,v in mins.items():
            #print(k,'\t',len(v[0]),len(v[1]),'\t',len(set(v[0].keys())-set(v[1].keys()))+len(set(v[1].keys())-set(v[0].keys())))

            for idx,a_min_dict in enumerate(v):
                mets_synthesis.append(k.id)
                # get metabolites
                try:
                    m=gem.metabolites.get_by_id(k.id)
                    met_description.append(m.name)
                except:
                    met_description.append(k)

                alternatives.append('alternative_%s'%str(idx+1))
                ## now we want to find up_regulated rxns and down_regulated reactions
                without_GPR=set([key.id for key in a_min_dict.keys()])
                rxns_in_min=set([key.id for key in a_min_dict.keys()]) & total_rxns
                de_in_min=rxns_in_min & set(de_rxns)
                percent_de.append(len(de_in_min)/len(rxns_in_min))
                rxns_num.append(len(rxns_in_min))
                list_de_in_min.append(','.join(de_in_min))
                list_rxn_in_min.append(','.join(rxns_in_min))

                without_gpr_rxns.append(len(without_GPR))
                ## total up and down reactions
                total_de_rxns=set(de_rxns)
                total_nochange_rxns=total_rxns-total_de_rxns
                ## apply hypergeometric function
                #M=len(total_rxns) ## number of gpr associated reactions (this can be true)
                M=len(model_rxns) ## number of gpr associated reactions
                n=len(de_rxns) ## total number of de regulated reactions
                N=len(rxns_in_min) ## number of reactions in a min
                x=len(de_in_min) ## number of deregulated in a min
                prob = hypergeom.cdf(x, M, n, N)

                probabilities.append(prob)
                ## calculate p-value
                pvalues.append(1-prob)


        df=pd.DataFrame()
        df['Metabolic task id']=mets_synthesis
        df['Alternative networks']=alternatives
        df['Metabolic task desciption']=met_description
        df[de_col+' rxns(%)']=percent_de
        df['GPR rxns']=rxns_num
        df['total # of rxns']=without_gpr_rxns
        df['p_values']=pvalues
        df['fdr_bh']=mtest.multipletests(pvalues, alpha=0.05, method='fdr_bh')[1]
        df[de_col+' rxns']=list_de_in_min
        df['rxns in a min']=list_rxn_in_min

        ## sort dataframe with p-values

        res_df=df.sort_values(by=['p_values'],ascending=True)
        res_df.to_csv(os.getcwd()+'/'+de_col+'_rxns.txt',sep='\t')



    def get_enrichment_table(self,mins,up_rxns,down_rxns,up_col='upregulated', down_col='downregulated'):
        ''' up and down can be high and low '''
        df=pd.DataFrame()
        mets_synthesis=[]
        alternatives=[]
        met_description=[]
        pvalues=[]
        probabilities=[]
        percent_up=[]
        percent_down=[]
        rxns_num=[]
        list_up_in_min=[]
        list_down_in_min=[]
        list_rxn_in_min=[]
        gem=self._source_gem
        total_rxns=[]
        model_rxns=[]
        for rxn in gem.reactions:
            model_rxns.append(rxn.id)
            if rxn.gene_reaction_rule!="":
                total_rxns.append(rxn.id)
        model_rxns=set(model_rxns)
        total_rxns=set(total_rxns)
        without_gpr_rxns=[]
        for k,v in mins.items():
            #print(k,'\t',len(v[0]),len(v[1]),'\t',len(set(v[0].keys())-set(v[1].keys()))+len(set(v[1].keys())-set(v[0].keys())))

            for idx,a_min_dict in enumerate(v):
                mets_synthesis.append(k.id)
                # get metabolites
                try:
                    m=gem.metabolites.get_by_id(k.id)
                    met_description.append(m.name)
                except:
                    met_description.append(k)

                alternatives.append('alternative_%s'%str(idx+1))
                ## now we want to find up_regulated rxns and down_regulated reactions
                without_GPR=set([key.id for key in a_min_dict.keys()])
                rxns_in_min=set([key.id for key in a_min_dict.keys()]) & total_rxns
                up_in_min=rxns_in_min & set(up_rxns)
                down_in_min=rxns_in_min & set(down_rxns)
                #nochange_in_min=rxns_in_min-up_in_min-down_in_min
                nochange_in_min=without_GPR-up_in_min-down_in_min
                # test division by error
                if len(rxns_in_min)>0:
                    percent_up.append((len(up_in_min)/len(rxns_in_min))*100)
                    percent_down.append((len(down_in_min)/len(rxns_in_min))*100)
                else:
                    percent_up.append(0)
                    percent_down.append(0)
                rxns_num.append(len(rxns_in_min))
                list_up_in_min.append(','.join(up_in_min))
                list_down_in_min.append(','.join(down_in_min))
                list_rxn_in_min.append(','.join(rxns_in_min))
                without_gpr_rxns.append(len(without_GPR))
                ## total up and down reactions
                total_up_rxns=set(up_rxns)
                total_down_rxns=set(down_rxns)
                #total_nochange_rxns=total_rxns-total_up_rxns-total_down_rxns ## this can be true
                total_nochange_rxns=total_rxns-total_up_rxns-total_down_rxns
                m=[len(total_up_rxns),len(total_down_rxns),len(total_nochange_rxns)];
                n=[len(up_in_min),len(down_in_min),len(nochange_in_min)];
                prob=mhygepdf(m,n);
                probabilities.append(prob)
                ## calculate p-value

                pval=0;
                for it1 in range(len(up_in_min),min(len(total_up_rxns),len(rxns_in_min))):
                    for it2 in range(len(down_in_min)):
                        if (it1+it2<len(rxns_in_min)):
                            pval=pval+mhygepdf(m,[it1,it2,len(rxns_in_min)-it1-it2]);
                pvalues.append(pval)
        df['Metabolic task id']=mets_synthesis
        df['Alternative networks']=alternatives
        df['Metabolic task desciption']=met_description
        df[up_col+' rxns(%)']=percent_up
        df[down_col+ ' rxns(%) ']=percent_down
        df['GPR rxns']=rxns_num
        df['total # of rxns']=without_gpr_rxns
        df['p_values']=pvalues
        df['fdr_bh']=mtest.multipletests(pvalues, alpha=0.05, method='fdr_bh')[1]
        df[up_col+' rxns']=list_up_in_min
        df[down_col+ ' rxns']=list_down_in_min
        df['rxns in a min']=list_rxn_in_min

        ## sort dataframe with p-values
        res_df=df.sort_values(by=['p_values'],ascending=True)
        res_df.to_csv(os.getcwd()+'/'+up_col+'_rxns.txt',sep='\t')





    def apply_enrichment(self,mins):
        ''' apply enrichment analysis '''
        params_rxn=self._params_rxn
        if ('up_rxns' in params_rxn.keys()) or ('down_rxns' in params_rxn.keys()):
            if (len(params_rxn['up_rxns'])>0) or (len(params_rxn['down_rxns'])>0):
                self.get_enrichment_table(mins,params_rxn['up_rxns'],params_rxn['down_rxns'],up_col='upregulated', down_col='downregulated')
                self.get_enrichment_table(mins,params_rxn['down_rxns'],params_rxn['up_rxns'],up_col='downregulated', down_col='upregulated')
                de_rxns=params_rxn['up_rxns']+params_rxn['down_rxns']
                self.get_enrichment_table_for_de(mins,de_rxns,de_col='deregulated')
            else:
                self.logger.info("empty up and down regulated reactions")
        elif ('high_rxns' in params_rxn.keys()) or ('low_rxns' in params_rxn.keys()):
            if (len(params_rxn['high_rxns'])>0) or (len(params_rxn['low_rxns'])>0):
                self.get_enrichment_table(mins,params_rxn['high_rxns'],params_rxn['low_rxns'],up_col='high_expressed', down_col='low_expressed')
            else:
                self.logger.info("empty up and down regulated reactions")
        else:
            self.logger.info("you need to reaction parameter for enrichment analysis 1) for condition change use up_ and down_rxns 2) for context use high_ and low_rxns")







    def run(self):
        # Extracting parameters

        biomass_rxn_ids = self.params["biomass_rxns"]
        biomass_rxns = [self._gem.reactions.get_by_id(x) for x in biomass_rxn_ids]
        main_bio_rxn = biomass_rxns[0]

        growth_rate = self.params["growth_rate"]
        cofactor_pairs = self.params["cofactor_pairs"]
        # Flatten cofactor_pairs list
        cofactors = [cofactor for pair in cofactor_pairs for cofactor in pair]
        lump_method = self.params["lump_method"]

        force_solve = self.params["force_solve"]
        timeout = self.params["timeout"]
        self._gem.solver.configuration.tolerances.feasibility = self.params["feasibility"]
        self._gem.solver.configuration.tolerances.integrality = self.params["feasibility"]
        self._source_gem.solver.configuration.tolerances.feasibility = self.params["feasibility"]
        self._source_gem.solver.configuration.tolerances.integrality = self.params["feasibility"]

        #### core reactions
        self.params['model_type']=self._model_type

        self.logger.info("Enumerating minmal networks ...")
        mt_model = EnumerateTask(self._source_gem, self.params) #### metabolic task model

        mins = mt_model.compute_mins(force_solve, method = lump_method)

        self.logger.info("Enumeration is done. now save the result in pickle file")
        pickle.dump(mins,open(os.getcwd()+'/mins.pickle','wb'))
        self.logger.info("Enumerating minimal networks ...")
        self.apply_enrichment(mins)
        # for k,v in mins.items():
        #     print(k,'\t',len(v[0]),len(v[1]),'\t',len(set(v[0].keys())-set(v[1].keys()))+len(set(v[1].keys())-set(v[0].keys())))




""" Minimum network enrichment analysis
.. moduleauthor:: vikash pandey
"""



from setuptools import setup, find_packages
# import os
# from pip.req import parse_requirements
# from pip.download import PipSession

# __location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
#
# def read_requirements():
#     '''parses requirements from requirements.txt'''
#     reqs_path = os.path.join(__location__, 'requirements.txt')
#     install_reqs = parse_requirements(reqs_path, session=PipSession())
#     reqs = [str(ir.req) for ir in install_reqs]
#     return reqs


version_tag = '0.0.1'

setup(name='mineapy',
      version=version_tag,
      author='vikash pandey',
      author_email='vikash.pandey@umu.se',
      url='https://github.com/vpandey-om/mineapy/',
      download_url='https://github.com/vpandey-om/mineapy/archive/v'+version_tag+'-alpha.tar.gz',
      install_requires=['cobra>0.13',
                        'bokeh>=0.12.1',
                        'networkx',
                        'optlang',
                        'pytest',
                        'scipy',
                        'tqdm',
                        'pytfa',
                        'statsmodels',
                        'sympy==1.6.1'],
      packages = find_packages(),
      python_requires='>=3.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
      description='mineapy, Minimum network enrichment analysis in Python',
      keywords=['mineapy','minea','thermodynamics','enrichment analysis'],

      license='Apache 2.0',

      # See https://PyPI.python.org/PyPI?%3Aaction=list_classifiers
      classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Environment :: Console',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: Apache Software License',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
      ],
     )



# -*- coding: utf-8 -*-
"""
.. module::
   :platform: Unix, Windows
   :synopsis: flux balance models accounting for expression, thermodynamics, and resource allocation constraints

.. moduleauthor::

Variables declarations

"""

from pytfa.optim.variables import GenericVariable, BinaryVariable, \
    ReactionVariable, ModelVariable, GeneVariable, get_binary_type, \
    ForwardUseVariable,BackwardUseVariable



class FluxSumVar(ModelVariable):
    """
    Class to represent a product A*B when performing linearization of the
    model
    """
    prefix = 'FSV_'



###  this function to calculate probability density function for multivariate hypergeometric function
import numpy as np
import scipy.special as sc
def mhygepdf(m,n):
    #MHYGEPDF Multivariate hypergeometric probability density function.
    # X = MHYGEPDF(M,N) returns the multivariate hypergeometric probability
    # density function at M with integer parameters in N. Note: The density
    # function is zero unless the elements in M are integers.
    #
    # If there are m_i elements of class i in a population and you take n
    # elements at random without replacement, then the number of elements of
    # each class in the sample (x_1,x_2,...,x_c) has the multivariate
    # hypergeometric distribution. This has the same relationship to the
    # multinomial distribution that the hypergeometric distribution has to the
    # binomial distribution--the multinomial distribution is the
    # 'with-replacement' distribution and the multivariate hypergeometric is
    # the 'without-replacement' distribution.
    #
    # Syntax: function x = mhygepdf(m,n)
    #
    # Inputs:
    # m - list of total (population) elements of class i
    # n - list of sampled elements of class i
    #
    # Output:
    # x - multivariate hypergeometric probability for values x_1,x_2,...,x_c
    #
    # Example. From the example given on the web Wikipedia: The Free
    #  Encyclopedia [http://en.wikipedia.org/wiki/Hypergeometric_distribution].
    #  Suppose there are 5 black, 10 white, and 15 red marbles in an urn. You
    #  reach in and randomly select six marbles without replacement. What is
    #  the probability that you pick exactly two of each color?
    #
    #             m = [5,10,15]; n = [2,2,2];
    #
    # Calling on Matlab the function:
    #             x = mhygepdf(m,n)
    #
    # Answer is: (in format long)
    #
    # x =
    #
    #   0.07957559681698
    # Creted by vikash Pandey

    ##### test 'm and must be a list of non-negative and integers.'

    m_int=[isinstance(item, int) for item in m ]
    n_int=[isinstance(item, int) for item in n ]
    if (not all(m_int)):
        print ('M must be a list of non-negative and integers.')
        exit()

    if (not all(n_int)):
        print ('N must be a list of non-negative and integers.')
        exit()
    ### We are going to multivariate compute hypergeometric pdf

    M = np.sum(m)
    N = np.sum(n)
    o = len(m)

    ###
    A = []
    for i in range(o):
        a = np.exp(sc.gammaln(m[i] + 1) - sc.gammaln(n[i] +1) - sc.gammaln(m[i] - n[i] + 1));
        A.append(a)
    ###
    B = np.exp(sc.gammaln(M + 1) - sc.gammaln(N +1) - sc.gammaln(M - N + 1));

    x = max(np.cumprod(A))/B;

    return x
